// jthz backup

#include <vcl.h>
#pragma hdrstop
#include <iostream.h>
#include <inifiles.hpp>
#include <stdlib.h> // exports atexit

void cleanup();
int error(HWND, PCSTR);
bool DirectoryExists(String);
void ForceDirectories(String);
bool maskMatch(String,String);
bool match(const char *, const char *);
void sweep(String, bool, void (*) (String, TSearchRec, bool));
void p_new(String, TSearchRec, bool);
void p_obs(String, TSearchRec, bool);
void p_dirKiller(String, TSearchRec, bool);
bool skip(String);
bool keepAnyway(String);

const int FOUND_SUCCESS = 0;
const String INI_SECTION = "jthz";

TStringList *recursed = 0, *non_recursed = 0, *include = 0;
TStringList *exclude = 0, *dirHitList = 0, *keep = 0;
String destination;

void cleanup(){
    // atexit handler
    delete include;
    include = 0;
    delete exclude;
    exclude = 0;
    delete recursed;
    recursed = 0;
    delete non_recursed;
    non_recursed = 0;
    delete keep;
    keep = 0;
}

int error(HWND hWnd, PCSTR s){
    MessageBox(hWnd, s, "SORRY", MB_ICONINFORMATION);
    return 0;
}

bool DirectoryExists(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    int Code = GetFileAttributes(Dir.c_str());
    return (Code != -1) && (FILE_ATTRIBUTE_DIRECTORY & Code);
}

void ForceDirectories(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    if(!Dir.Length())
        return;
    if(*AnsiLastChar(Dir) == '\\')
        Dir.Delete(Dir.Length(), 1);
    if((Dir.Length() < 3) || DirectoryExists(Dir) ||
            (ExtractFilePath(Dir) == Dir))
        return;
    ForceDirectories(ExtractFilePath(Dir));
    CreateDir(Dir);
}

bool maskMatch(String mask, String fileName){
    // returns true if fileName matches standard dos wildcard mask
    // derived from Bob Stout's XSTRCMP.C
    return match(mask.LowerCase().c_str(),
            fileName.LowerCase().c_str());
}

bool match(const char *mask,const char *s){
    // recursive helperfunction for maskMatch
    switch(*mask){
        case '\0':
            return !*s;
        case '*':
            return match(mask + 1, s) || *s && match(mask, s + 1);
        case '?':
            return *s && match(mask + 1, s + 1);
    }
    return(*mask == *s) && match(mask + 1, s + 1);
}

void sweep(String inputDir, bool bRecurse,
        void (*generic) (String, TSearchRec, bool)){
    // recurses directory if bRecurse==true
    if(!inputDir.Length())
        return;
    if(*AnsiLastChar(inputDir) != '\\')
        inputDir += "\\";
    TSearchRec sr;
    if(FindFirst(inputDir + "*.*", faAnyFile, sr) == FOUND_SUCCESS){
        generic(inputDir, sr, bRecurse);
        while(FindNext(sr) == FOUND_SUCCESS)
            generic(inputDir, sr, bRecurse);
        FindClose(sr);
    }
}

void p_new(String path, TSearchRec tsr, bool bRecurse){
    // sweep plugin-function to hunt for new files
    if(tsr.Attr & faDirectory){
        if(bRecurse && tsr.Name[1] != '.')
            sweep(path + tsr.Name, true, p_new);
    }
    else if(!skip(tsr.Name)){
        // build copy-destination (global 'destination' ends with \)
        String destFile = destination + path[1] + "\\" +
                path.SubString(4, path.Length() - 3) + tsr.Name;
        TSearchRec srDest;
        if(FindFirst(destFile, faAnyFile, srDest) == FOUND_SUCCESS){
            // file exists; is our file newer?
            // use y2k-proof 64bit arithmetic     =8)~
            LARGE_INTEGER sourceTimeDate, destTimeDate;
            CopyMemory(&sourceTimeDate, &tsr.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            CopyMemory(&destTimeDate, &srDest.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            if(sourceTimeDate.QuadPart > destTimeDate.QuadPart){
                cout << path.c_str() << tsr.Name.c_str() << endl;
                cout << "File is updated - copying..." << endl;
                // remove RHSA
                SetFileAttributes(destFile.c_str(),
                        FILE_ATTRIBUTE_NORMAL);
                CopyFile((path + tsr.Name).c_str(),
                        destFile.c_str(), FALSE);
            }
            FindClose(srDest);
        }
        else{
            cout << path.c_str() << tsr.Name.c_str() << endl;
            cout << "File is new - copying..." << endl;
            ForceDirectories(ExtractFilePath(destFile));
            CopyFile((path + tsr.Name).c_str(), destFile.c_str(), FALSE);
        }
    }
}

void p_obs(String path, TSearchRec tsr, bool bRecurse){
    // sweep plugin-function to hunt for obsolete files
    if(tsr.Attr & faDirectory){
        if(bRecurse && tsr.Name[1] != '.')
            sweep(path + tsr.Name, true, p_obs);
    }
    else{
        // map destination-path back to source-path
        String destFile = path + tsr.Name;
        int len = destination.Length();
        String checkFile = destFile.SubString(len, destFile.Length() - len + 1);
        checkFile[1] = destFile[len + 1];
        checkFile[2] = ':';
        checkFile[3] = '\\';
        if(!FileExists(checkFile) || skip(tsr.Name)){
            if(!keepAnyway(tsr.Name)){
                // remove RHSA
                SetFileAttributes(destFile.c_str(), FILE_ATTRIBUTE_NORMAL);
                DeleteFile(destFile.c_str());
            }
        }
    }
}

void p_dirKiller(String path, TSearchRec tsr, bool bRecurse){
    // sweep plugin-function to hunt for obsolete directories
    if(tsr.Attr & faDirectory){
        if(bRecurse&&tsr.Name[1] != '.'){
            dirHitList->Add(path + tsr.Name);
            sweep(path + tsr.Name, true, p_dirKiller);
        }
    }
}

bool skip(String fileName){
    // returns true if fileName should be skipped
    // based on entries in global TStringLists exclude and include
    // standard dos wildcards * and ? apply
    if(!fileName.Length())
        return true; // skip
    for(int i = 0; i < exclude->Count; ++i)
        if(maskMatch(exclude->Strings[i], fileName))
            return true; // skip
    if(include->Count){ // file must be included here to pass
        for(int i = 0; i < include->Count; ++i)
            if(maskMatch(include->Strings[i], fileName))
                return false; // go ahead
        return true; // skip
    }
    return false; // go ahead
}

bool keepAnyway(String fileName){
    // returns true if fileName should be skipped (not obsolete deleted)
    // based on entries in global TStringList keep
    // standard dos wildcards * and ? apply
    if(!fileName.Length())
        return true; // skip
    for(int i = 0; i < keep->Count; ++i)
        if(maskMatch(keep->Strings[i], fileName))
            return true; // keep anyway
    return false; // nope, throw away
}

int main(int argc, char* argv[]){
    char szTitle[256];
    HWND hConsole = NULL;
    // get our own handle
    if(GetConsoleTitle(szTitle, sizeof(szTitle)))
        hConsole = FindWindow(NULL, szTitle);
    // see if we are started with a cmdline parm
    if(argc != 2)
        return error(hConsole, "Run me with a parameter");
    // try to read info from file
    recursed = new TStringList(); // recurse these directories
    non_recursed = new TStringList(); // just scan these directories
    exclude = new TStringList(); // exclude files matchin these masks
    include = new TStringList(); // include files matchin these masks
    keep = new TStringList(); // include files matchin these masks
    TIniFile *ini = new TIniFile(argv[1]);
    recursed->CommaText = ini->ReadString(INI_SECTION, "recursed", "");
    non_recursed->CommaText = ini->ReadString(INI_SECTION, "non-recursed", "");
    destination = ini->ReadString(INI_SECTION, "destiny", "");
    exclude->CommaText = ini->ReadString(INI_SECTION, "xclude", "");
    include->CommaText = ini->ReadString(INI_SECTION, "include", "");
    keep->CommaText = ini->ReadString(INI_SECTION, "keep on destiny", "");
    delete ini;
    // install cleaner-handler
    atexit(cleanup);
    // check validity of input
    if(!destination.Length())
        return error(hConsole, "No destination scanned");
    if(*AnsiLastChar(destination) != '\\')
        destination += "\\";
    // make sure destination exists
    ForceDirectories(destination);
    // now see if we have access to destination
    if(!DirectoryExists(destination))
        return error(hConsole, "Invalid destination");
    // source(s) valid?
    if(!recursed->Count && !non_recursed->Count)
        return error(hConsole, "No source-directories scanned");
    for(int i = 0; i < recursed->Count; ++i)
        if(!DirectoryExists(recursed->Strings[i]))
            return error(hConsole, "Non-existing directory in recursed dirs");
    for(int i = 0; i < non_recursed->Count; ++i)
        if(!DirectoryExists(non_recursed->Strings[i]))
            return error(hConsole, "Non-existing directory in non-recursed entry");
    // first round; copy new files from source to destination
    cout << "Hunting down new files..." << endl << endl;
    for(int i = 0; i < recursed->Count; ++i)
        sweep(recursed->Strings[i], true, p_new);
    for(int i = 0; i < non_recursed->Count; ++i)
        sweep(non_recursed->Strings[i], false, p_new);
    // second round; remove obsolete files
    cout << "Hunting down obsolete files..." << endl << endl;
    for(int i = 0; i < recursed->Count; ++i){
        String tmp = recursed->Strings[i];
        if(!tmp.Length())
            continue;
        sweep(destination + tmp[1] + "\\" +
                tmp.SubString(4, tmp.Length() - 3), true, p_obs);
    }
    for(int i = 0; i < non_recursed->Count; ++i){
        String tmp = non_recursed->Strings[i];
        if(!tmp.Length())
            continue;
        sweep(destination + tmp[1] + "\\" +
                tmp.SubString(4, tmp.Length() - 3), false, p_obs);
    }
    // third round; remove obsolete directories resulting from round two
    cout << "Hunting down obsolete directories..." << endl << endl;
    dirHitList = new TStringList(); // create container
    // gather all directories in destination
    sweep(destination, true, p_dirKiller);
    // kill them all.. that'll distract 'em
    for(int i = 0; i < dirHitList->Count; ++i){
        String tmp = dirHitList->Strings[i];
        while(RemoveDir(tmp)){ // failure indicates dir ain't empty
            tmp = ExtractFilePath(tmp);
            if(tmp.Length() && *AnsiLastChar(tmp) == '\\')
                tmp.Delete(tmp.Length(), 1);
        }
    }
    delete dirHitList;
    return 0;
}
